#ifndef COMMON_H
#include "common.h"
#endif

int set_flowlabel(int sock, struct sockaddr_in6 *dst, __u32 label)
{
    int on = 1;
    struct in6_flowlabel_req freq;

    memset(&freq, 0, sizeof(freq));
    freq.flr_label = htonl(label);
    freq.flr_action = IPV6_FL_A_GET;
    freq.flr_flags = IPV6_FL_F_CREATE;
    freq.flr_share = IPV6_FL_S_EXCL;
    memcpy(&freq.flr_dst, &dst->sin6_addr, 16);
    if (setsockopt(sock, SOL_IPV6, IPV6_FLOWLABEL_MGR, &freq, sizeof(freq)) == -1) {
	perror("can't lease flowlabel");
	return -1;
    }
// if next line was comment-out, flowlabel in packet didn't appear.
    dst->sin6_flowinfo = freq.flr_label;
    if (setsockopt(sock, SOL_IPV6, IPV6_FLOWINFO_SEND, &on, sizeof(on)) == -1) {
	perror("can't send flowinfo");
// release a flowlabel that was given by FLOWLABEL_MGR
	freq.flr_action = IPV6_FL_A_PUT;
	setsockopt(sock, SOL_IPV6, IPV6_FLOWLABEL_MGR, &freq, sizeof(freq));
	dst->sin6_flowinfo = 0;
	return -1;
    }
    return 0;
}

void release_flowlabel(int sock, struct sockaddr_in6 *dst, __u32 label)
{
    struct in6_flowlabel_req freq;

    memset(&freq, 0, sizeof(freq));
    freq.flr_label = htonl(label);
    freq.flr_action = IPV6_FL_A_PUT;
    freq.flr_flags = IPV6_FL_F_CREATE;
    freq.flr_share = IPV6_FL_S_EXCL;
    memcpy(&freq.flr_dst, &dst->sin6_addr, 16);
    setsockopt(sock, SOL_IPV6, IPV6_FLOWLABEL_MGR, &freq, sizeof(freq));
// if value of flowinfo was remained, Invalid argument error was occured.
    dst->sin6_flowinfo = 0;
}

int set_rxflow(int sock)
{
    int on = 1;

#ifdef USE_PKTINF
// return dst_addr of packet (= own address)
    if (setsockopt(sock, SOL_IPV6, IPV6_PKTINFO, &on, sizeof(on)) == -1) {
	perror("can't get address");
	return -1;
    }
#endif
    if (setsockopt(sock, SOL_IPV6, IPV6_FLOWINFO, &on, sizeof(on)) == -1) {
	perror("can't get flowinfo");
	return -1;
    }
    return 0;
}

size_t recvlabel(int sock, char *buf, size_t len, int flags,
		struct sockaddr *addr, int *addrlen)
{
    size_t cc;
    char cbuf[128];
    struct cmsghdr *c;
    struct iovec iov = { buf, len };
    struct msghdr msg = { addr, *addrlen, &iov, 1, cbuf, sizeof(cbuf), 0 };

    cc = recvmsg(sock, &msg, flags);
    if (cc < 0) return cc;
    ((struct sockaddr_in6*)addr)->sin6_flowinfo = 0;
    *addrlen = msg.msg_namelen;
    for (c=CMSG_FIRSTHDR(&msg); c; c = CMSG_NXTHDR(&msg, c)) {
	printf("level = %d, type = %d\n",c->cmsg_level, c->cmsg_type);
	if (c->cmsg_level == SOL_IPV6) {
	    if (c->cmsg_type == IPV6_FLOWINFO) {
		((struct sockaddr_in6*)addr)->sin6_flowinfo = *(__u32*)CMSG_DATA(c);
#ifdef USE_PKTINF
	    } else if (c->cmsg_type == IPV6_PKTINFO) {
		memcpy(&((struct sockaddr_in6*)addr)->sin6_addr, CMSG_DATA(c), 16);
#endif
	    }
	}
    }
    return cc;
}

int getip6name(struct in6_addr *addr, char *host, char *kind)
{
    char ip6addr[MAX_NAME_CHAR];
    int hlen;

#ifdef SELECT_HBN2
    if (inet_pton(AF_INET6, host, addr) <= 0) {
	struct hostent *hp;

	hp = gethostbyname2(host, AF_INET6);
	if (hp == NULL) {
	    fprintf(stderr, "unknown host %s\n", host);
	    return -3;
	}
	memcpy(addr, hp->h_addr_list[0], 16);
    }
#else
    {
	struct addrinfo hints, *res;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET6;
	hlen = getaddrinfo(host, NULL, &hints, &res);
	if (hlen) {
	    fprintf(stderr, "getaddrinfo : %s\n", gai_strerror(hlen));
	    return -3;
	}
	memcpy(addr, res->ai_addr, res->ai_addrlen);
    }
#endif

    memset(ip6addr, 0, MAX_NAME_CHAR);
    hlen = (strlen(host) < MAX_NAME_CHAR -1 ? strlen(host) : MAX_NAME_CHAR-1);
    if (inet_ntop(AF_INET6, addr, ip6addr, 128) == NULL) {
	perror("inet_ntop");
	memcpy(ip6addr, host, hlen);
    }
    printf("%s = %s\n",kind, ip6addr);

    return 0;
}

size_t getlabel(int sock, struct sockaddr_in6 *addr, int *addrlen)
{
    size_t cc;
    char cbuf[128];
    struct cmsghdr *c;
    struct msghdr msg = {0};
    int cbuflen;

    cbuflen = sizeof(cbuf);
    msg.msg_control = cbuf;
    msg.msg_controllen = cbuflen;
    cc = getsockopt(sock, SOL_IPV6, IPV6_PKTOPTIONS, cbuf, &cbuflen);
    if (cc < 0) return cc;
    (addr)->sin6_flowinfo = 0;
    *addrlen = cbuflen;
    msg.msg_controllen = cbuflen;
    for (c=CMSG_FIRSTHDR(&msg); c; c = CMSG_NXTHDR(&msg, c)) {
	printf("level = %d, type = %d\n",c->cmsg_level, c->cmsg_type);
	if (c->cmsg_level == SOL_IPV6) {
	    if (c->cmsg_type == IPV6_FLOWINFO) {
		(addr)->sin6_flowinfo = *(__u32*)CMSG_DATA(c);
#ifdef USE_PKTINF
	    } else if (c->cmsg_type == IPV6_PKTINFO) {
		memcpy(&(addr)->sin6_addr, CMSG_DATA(c), 16);
#endif
	    }
	}
    }
    return cc;
}

size_t setlabel(int sock, __u32 flowinfo)
{
    size_t cc;
    char cbuf[128];
    struct cmsghdr *c;
    struct msghdr msg;
    int cbuflen;

    memset(&msg, 0, sizeof(msg));
    memset(cbuf, 0, cbuflen);
    msg.msg_control = cbuf;
    msg.msg_controllen = sizeof(cbuf);
    c = CMSG_FIRSTHDR(&msg);
    c->cmsg_level = SOL_IPV6;
    c->cmsg_type = IPV6_FLOWINFO;
    c->cmsg_len = CMSG_LEN(sizeof(__u32));
    memcpy(CMSG_DATA(c), &flowinfo, sizeof(__u32));
//    printf("flowlabel = 0x%x\n",ntohl(*(__u32 *)CMSG_DATA(c)));
    cbuflen = c->cmsg_len;
    cc = setsockopt(sock, SOL_IPV6, IPV6_PKTOPTIONS, cbuf, cbuflen);
    return cc;
}
